#![allow(non_snake_case)]
#![allow(dead_code)]

use std::ffi::CString;
use std::mem::size_of;
use std::os::raw::c_void;
use std::ptr;

use cgmath::{ Vector3, Vector2 };
use cgmath::prelude::*;
use glow::*;
use field_offset::offset_of;

use crate::shader::shader;
// NOTE: without repr(C) the compiler may reorder the fields or use different padding/alignment than C.
// Depending on how you pass the data to OpenGL, this may be bad. In this case it's not strictly
// necessary though because of the `offset!` macro used below in setupMesh()
#[repr(C)]
pub struct Vertex {
    // position
    pub Position: Vector3<f32>,
    // normal
    pub Normal: Vector3<f32>,
    // texCoords
    pub TexCoords: Vector2<f32>,
    // tangent
    pub Tangent: Vector3<f32>,
    // bitangent
    pub Bitangent: Vector3<f32>,
}

impl Default for Vertex {
    fn default() -> Self {
        Vertex {
            Position: Vector3::zero(),
            Normal: Vector3::zero(),
            TexCoords: Vector2::zero(),
            Tangent: Vector3::zero(),
            Bitangent: Vector3::zero(),
        }
    }
}

#[derive(Clone)]
pub struct Texture {
    pub id: glow::NativeTexture,
    pub type_: String,
    pub path: String,
}

pub struct Mesh {
    /*  Mesh Data  */
    pub vertices: Vec<Vertex>,
    pub indices: Vec<u32>,
    pub textures: Vec<Texture>,
    pub VAO: glow::VertexArray,

    /*  Render data  */
    VBO: glow::Buffer,
    EBO: glow::Buffer,
}

impl Mesh {
    pub fn new(vertices: Vec<Vertex>, indices: Vec<u32>, textures: Vec<Texture>,gl: std::rc::Rc<glow::Context>) -> Mesh {

        let mut mesh =
            {
        unsafe {
        let mut mesh = Mesh {
            vertices, indices, textures,
            VAO: gl.create_vertex_array().unwrap(), VBO: gl.create_buffer().unwrap(), EBO: gl.create_buffer().unwrap()
        };
                mesh
        }
            };
        // now that we have all the required data, set the vertex buffers and its attribute pointers.
        unsafe { mesh.setupMesh(gl) }
        mesh
    }

    /// render the mesh
    pub unsafe fn Draw(&self, shader: &shader) {
        // bind appropriate textures
        let mut diffuseNr  = 0;
        let mut specularNr = 0;
        let mut normalNr   = 0;
        let mut heightNr   = 0;
        for (i, texture) in self.textures.iter().enumerate() {
            shader.gl.active_texture(glow::TEXTURE0 + i as u32); // active proper texture unit before binding
            // retrieve texture number (the N in diffuse_textureN)
            let name = &texture.type_;
            let number = match name.as_str() {
                "texture_diffuse" => {
                    diffuseNr += 1;
                    diffuseNr
                },
                "texture_specular" => {
                    specularNr += 1;
                    specularNr
                }
                "texture_normal" => {
                    normalNr += 1;
                    normalNr
                }
                "texture_height" => {
                    heightNr += 1;
                    heightNr
                }
                _ => panic!("unknown texture type")
            };
            // now set the sampler to the correct texture unit
            let sampler = String::from(format!("{}{}", name, number));
            shader.setInt(&sampler, i as i32);
            // and finally bind the texture
            shader.gl.bind_texture(glow::TEXTURE_2D, Some(texture.id));
        }

        // draw mesh
        shader.gl.bind_vertex_array(Some(self.VAO));
        shader.gl.draw_elements(glow::TRIANGLES, self.indices.len() as i32, glow::UNSIGNED_INT, 0);
        shader.gl.bind_vertex_array(None);

        // always good practice to set everything back to defaults once configured.
        shader.gl.active_texture(glow::TEXTURE0);
    }

    unsafe fn setupMesh(&mut self, gl: std::rc::Rc<glow::Context>) {
        // create buffers/arrays
        gl.bind_vertex_array(Some(self.VAO));
        // load data into vertex buffers
        gl.bind_buffer(glow::ARRAY_BUFFER, Some(self.VBO));
        // A great thing about structs with repr(C) is that their memory layout is sequential for all its items.
        // The effect is that we can simply pass a pointer to the struct and it translates perfectly to a glm::vec3/2 array which
        // again translates to 3/2 floats which translates to a byte array.
        //let size = (self.vertices.len() * size_of::<Vertex>()) as isize;
        //let data = &self.vertices[0] as *const Vertex as *const c_void;
        let data: &[u8] = core::slice::from_raw_parts(
            self.vertices.as_ptr() as *const u8,
            self.vertices.len() * core::mem::size_of::<f32>(),
        );




        gl.buffer_data_u8_slice(glow::ARRAY_BUFFER, data, glow::STATIC_DRAW);

        gl.bind_buffer(glow::ELEMENT_ARRAY_BUFFER, Some(self.EBO));
        let data: &[u8] = core::slice::from_raw_parts(
            self.indices.as_ptr() as *const u8,
            self.indices.len() * core::mem::size_of::<f32>(),
        );
        gl.buffer_data_u8_slice(glow::ELEMENT_ARRAY_BUFFER, data, glow::STATIC_DRAW);

        // set the vertex attribute pointers
        let size = size_of::<Vertex>() as i32;
        // vertex Positions
        gl.enable_vertex_attrib_array(0);
        gl.vertex_attrib_pointer_f32(0, 3, glow::FLOAT, false, size, offset_of!(Vertex => Position).get_byte_offset() as i32);
        // vertex normals
        gl.enable_vertex_attrib_array(1);
        gl.vertex_attrib_pointer_f32(1, 3, glow::FLOAT, false, size, offset_of!(Vertex => Normal).get_byte_offset() as i32);
        // vertex texture coords
        gl.enable_vertex_attrib_array(2);
        gl.vertex_attrib_pointer_f32(2, 2, glow::FLOAT, false, size, offset_of!(Vertex => TexCoords).get_byte_offset() as i32);
        // vertex tangent
        gl.enable_vertex_attrib_array(3);
        gl.vertex_attrib_pointer_f32(3, 3, glow::FLOAT, false, size, offset_of!(Vertex => Tangent).get_byte_offset() as i32);
        // vertex bitangent
        gl.enable_vertex_attrib_array(4);
        gl.vertex_attrib_pointer_f32(4, 3, glow::FLOAT, false, size, offset_of!(Vertex => Bitangent).get_byte_offset() as i32);

        gl.bind_vertex_array(None);
    }
}
